@@ -15,6 +15,8 @@ def generate_fmu_eval(input_args, zero_assign_ptrs, output_ptrs, parameters_retu
     for arg_id in input_args:
         if arg_id not in fmu_output_args:
             args_lst.append(ast.arg(arg=arg_id))
+    args_lst.append(ast.arg(arg="_time"))
+
     args = ast.arguments(posonlyargs=[], args=args_lst, kwonlyargs=[],
                          kw_defaults=[], defaults=[])
 
@@ -35,7 +37,7 @@ def generate_fmu_eval(input_args, zero_assign_ptrs, output_ptrs, parameters_retu
         if arg_id not in fmu_output_args:
             eq_call_args.append(ast.Name(id=arg_id, ctx=ast.Load()))
 
-    eq_call_args.append(ast.Constant(value=0.1))
+    eq_call_args.append(ast.Name(id="_time", ctx=ast.Load()))
     eq_expr = [ast.Expr(value=ast.Call(func=ast.Name(id=EQ_CALL), args=eq_call_args, keywords=[], ctx=ast.Load()),
                         keywords=[])]
     return_elts = []
@@ -78,12 +80,8 @@ def add_address_as_void_pointer(var_id):
 
 def _generate_fet_real_expr():
     return ast.Expr(value=ast.Call(func=ast.Name(id='getreal', ctx=ast.Load()),
-                                   args=[ast.Name(id='component', ctx=ast.Load()),
-                                         ast.Attribute(value=ast.Name(id='vr', ctx=ast.Load()), attr='ctypes',
-                                                       ctx=ast.Load()),
-                                         ast.Name(id='len_q', ctx=ast.Load()),
-                                         ast.Attribute(value=ast.Name(id='value', ctx=ast.Load()), attr='ctypes',
-                                                       ctx=ast.Load())],
+                                   args=[ast.Name(id='component', ctx=ast.Load()), ast.Name(id='vr', ctx=ast.Load()),
+                                         ast.Name(id='len_q', ctx=ast.Load()), ast.Name(id='value', ctx=ast.Load())],
                                    keywords=[]),
                     keywords=[])
 
@@ -157,8 +155,7 @@ def generate_eval_llvm(assign_ptrs, output_args, states_idx, var_order: list, ou
 
         body.append(ast.Expr(value=ast.Call(func=ast.Name(id='fmi2SetC', ctx=ast.Load()),
                                             args=[ast.Name(id='component', ctx=ast.Load()),
-                                                  ast.Attribute(value=ast.Name(id='value3', ctx=ast.Load()),
-                                                                attr='ctypes', ctx=ast.Load()),
+                                                  ast.Name(id='value3', ctx=ast.Load()),
                                                   ast.Constant(value=len(states_idx))], keywords=[])))
 
     body.append(ast.Expr(value=ast.Call(func=ast.Name(id='set_time', ctx=ast.Load()),
@@ -247,9 +244,9 @@ def generate_eval_event(state_idx, len_q, var_order: list, event_id):
                                               ast.Name(id='t', ctx=ast.Load())], keywords=[])))
 
     body.append(ast.Expr(value=ast.Call(func=ast.Name(id='get_event_indicators', ctx=ast.Load()),
-                                        args=[ast.Name(id='component', ctx=ast.Load()), ast.Attribute(
-                                            value=ast.Name(id='value_event', ctx=ast.Load()), attr='ctypes',
-                                            ctx=ast.Load()), ast.Name(id='event_n', ctx=ast.Load())],
+                                        args=[ast.Name(id='component', ctx=ast.Load()),
+                                              ast.Name(id='value_event', ctx=ast.Load()),
+                                              ast.Name(id='event_n', ctx=ast.Load())],
                                         keywords=[])))
     lst_arg = []
     for i in range(len_q):
@@ -266,11 +263,9 @@ def generate_eval_event(state_idx, len_q, var_order: list, event_id):
 
     body.append(ast.Expr(value=ast.Call(func=ast.Name(id='fmi2SetReal', ctx=ast.Load()),
                                         args=[ast.Name(id='component', ctx=ast.Load()),
-                                              ast.Attribute(value=ast.Name(id='vr', ctx=ast.Load()),
-                                                            attr='ctypes', ctx=ast.Load()),
+                                              ast.Name(id='vr', ctx=ast.Load()),
                                               ast.Name(id='len_q', ctx=ast.Load()),
-                                              ast.Attribute(value=ast.Name(id='value2', ctx=ast.Load()),
-                                                            attr='ctypes', ctx=ast.Load())], keywords=[])),
+                                              ast.Name(id='value2', ctx=ast.Load())], keywords=[])),
                 )
 
     body.append(ast.Assign(targets=[ast.Subscript(
@@ -393,11 +388,9 @@ def generate_set_fmi_update(arg_elts):
 
     body.append(ast.Expr(value=ast.Call(func=ast.Name(id='fmi2SetReal', ctx=ast.Load()),
                                         args=[ast.Name(id='component', ctx=ast.Load()),
-                                              ast.Attribute(value=ast.Name(id='vr', ctx=ast.Load()),
-                                                            attr='ctypes', ctx=ast.Load()),
+                                              ast.Name(id='vr', ctx=ast.Load()),
                                               ast.Name(id='len_q', ctx=ast.Load()),
-                                              ast.Attribute(value=ast.Name(id='value1', ctx=ast.Load()),
-                                                            attr='ctypes', ctx=ast.Load())], keywords=[])))
+                                              ast.Name(id='value1', ctx=ast.Load())], keywords=[])))
     return body
 
 
@@ -458,13 +451,9 @@ def generate_action_event(len_q: int, var_order: list):
                                                                     ctx=ast.Load()))]), lineno=0))
     body.append(ast.Expr(
         value=ast.Call(func=ast.Name(id='getreal', ctx=ast.Load()), args=[ast.Name(id='component', ctx=ast.Load()),
-                                                                          ast.Attribute(
-                                                                              value=ast.Name(id='vr', ctx=ast.Load()),
-                                                                              attr='ctypes', ctx=ast.Load()),
+                                                                          ast.Name(id='vr', ctx=ast.Load()),
                                                                           ast.Name(id='len_q', ctx=ast.Load()),
-                                                                          ast.Attribute(value=ast.Name(id='value',
-                                                                                                       ctx=ast.Load()),
-                                                                                        attr='ctypes', ctx=ast.Load())],
+                                                                          ast.Name(id='value', ctx=ast.Load())],
                        keywords=[])))
 
     for i in range(len_q):
@@ -594,6 +583,7 @@ def generate_eq_call(deriv_names, var_names, input_var_names_ordered, output_var
     for v_name in var_names:
         if v_name not in output_var_names_ordered:
             args.append(ast.Attribute(value=ast.Name(id='scope', ctx=ast.Load()), attr=v_name, ctx=ast.Load()))
+    args.append(ast.Attribute(value=ast.Name(id='scope', ctx=ast.Load()), attr="global_vars.t", ctx=ast.Load()))
 
     return ast.Module(body=[ast.FunctionDef(name='eval',
                                             args=ast.arguments(posonlyargs=[],
